/* 
 * This program is written by 
 *               
 *                          Murtaza Amiji
 *		Department Of ELectrical & Computer Engineering,
 *		     Worcester Polytechnic Institute,
 *		            Worcester, MA.
 *                    E-Mail: murti@ece.wpi.edu
 *               Personal Homepage: http://www.wpi.edu/~murti
 *                    ECE Homepage: http://ece.wpi.edu
 *
 * Copyright (c) 1995-96 Worcester Polytechnic Institute. 
 * All rights reserved.
 * 
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for non commericial purpose, provided that 
 * the above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL WORCESTER POLYTECHNIC INSTITUTE BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF WORCESTER 
 * POLYTECHNIC INSTITUTE HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * WORCESTER POLYTECHNIC INSTITUTE SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND WORCESTER POLYTECHNIC INSTITUTE HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


#include "ephone.h"

void message(char *message)
{
  int i;
  for(i=0; message[i] != '\0' ;i++) {
    printf("%c", message[i]); 
    fflush(stdout);
  }
  printf("\n"); 
  fflush(stdout);
}
      
int open_recording()
{
  
  if((audio_rd = open(AUDIO, O_RDONLY, 0)) == -1) {
    close(audio_wr);
    close(audio_rd);
    if((audio_rd = open(AUDIO, O_RDONLY, 0)) == -1) {
      message("Cannot open /dev/dsp for recording"); 
      return (-1);         
    }
  }
  
  if((DSP_config()) < 0) {
    message("Error setting the sampling parameters.");
    return (-1);
  }   
  
  return 0;
}  

int open_playback()
{  
  
  if((audio_wr = open(AUDIO, O_WRONLY, 0)) == -1) {
    close(audio_wr);
    close(audio_rd);
    if((audio_wr = open(AUDIO, O_WRONLY, 0)) == -1) {
      message("Error: Cannot open /dev/dsp for playback");
      return (-1);
    }     
  }
  
  if((DSP_config()) < 0) {
    message("Error setting the sampling parameter for playback.");
    return (-1);
  }
  return 0;
}  

int threshold()
{
  int samples_read = 0;
  RECEIVING = FALSE;
  TRANSMITTING = TRUE;
  if((open_recording()) == -1) {
    message("Cannot open the recording device to set the cutoff amplitude");
    return (-1);
  }  
  message("Hold on..! Dont start talking.");
  message("Calculating the threshold value for the default audio buffer...");
  while(steps < 200) {
  samples_read = read(audio_rd, record_buffer, FRAGMENT);
  if((squelch_buffer(FRAGMENT)) == 0)
    total_average_amplitude+=average_amplitude;
  steps++;
  }
  close(audio_rd); 
  cutoff_amplitude = (float) total_average_amplitude/steps;
  cutoff_amplitude+=100; /* a slight adjustment from the calculated value */
  total_average_amplitude = 0;
  steps = 0;
  printf("cut-off amplitude is ==>> %f\n", cutoff_amplitude); 
  fflush(stdout);
  return 0;

}
  
int squelch_buffer(int samples_read)
{
 int total_amplitude, i;
  
 squelch = 0.963*squelch + 0.03*average_amplitude + 1; /* adaptive squelch value. */
  for(i=0;i<samples_read;i++)
    total_amplitude+= abs(record_buffer[i]);

  average_amplitude = (float) total_amplitude/samples_read;
  
  if(average_amplitude < squelch || average_amplitude < cutoff_amplitude) 
    count++;
  
  if(average_amplitude >= squelch && average_amplitude > cutoff_amplitude) {
    count--;
    if(count < 0)
      count = 0;
  }
 
  /* ditch the second consecutive silence buffer. */

  if(count == 15) {  
    count--; 
    return (-1);
  } 
 return 0;
}
 

int Init_Soundcard()
{  
  if((audio_rd = open(AUDIO, O_RDONLY, 0)) > 0) {
    message("Detected one half duplex sound card.");
    close(audio_rd);
    return 0;         
  } 
  
  message("Error: Program failed to detect a sound card.");
  return (-1);
  
}

int DSP_config()
{
  if(RECEIVING) {
    if(ioctl(audio_wr, SOUND_PCM_WRITE_BITS, &bits_per_sample) == -1) {
      message("Error: cannot set the half duplex card to play at 16bits/sample.");
      return (-1);
    }    
    
    if(ioctl(audio_wr, SNDCTL_DSP_SETFRAGMENT, &dma_buffer_size) == -1) {
      message("Error: cannot set the DMA buffer to play 16msec worth of audio.");
      return (-1);
    } 
  }

  else if(TRANSMITTING) { 
    if(ioctl(audio_rd, SOUND_PCM_WRITE_BITS, &bits_per_sample) == -1) {
      message("Error: cannot set the half duplex card to record at 16bits/sample");
      return (-1);
    }    
    
    if(ioctl(audio_rd, SNDCTL_DSP_SETFRAGMENT, &dma_buffer_size) == -1) {
      message("Error: cannot set the DMA buffer to record 16msec worth of audio.");
      return (-1);
    } 
  }
  return 0;      

}
  
int isEmpty(QUEUENODEPTR headPtr)
{
   return headPtr == NULL;
} 
 
void enqueue(QUEUENODEPTR *headPtr, QUEUENODEPTR *tailPtr)
{
   QUEUENODEPTR newPtr;
   newPtr = malloc(sizeof(QUEUENODE));
   if(newPtr != NULL) {
     memcpy(newPtr->data, receive_buffer, compressed_size);
      newPtr->nextPtr = NULL;
     if(isEmpty(*headPtr))
        *headPtr = newPtr;     
     else
       (*tailPtr)->nextPtr = newPtr;
       *tailPtr = newPtr;
   }
   else
      message("Memory exhausted. Cannot add any more audio buffers.");
}

void dequeue(QUEUENODEPTR *headPtr, QUEUENODEPTR *tailPtr)
{
   
   QUEUENODEPTR tempPtr;

   memcpy(compressed_buffer, (*headPtr)->data, compressed_size);
   
   tempPtr = *headPtr;
   *headPtr = (*headPtr)->nextPtr;
 
   if (*headPtr == NULL)
	*tailPtr = NULL;
      free(tempPtr);
  
}

int network_connection()
{
  char response_buffer[1024];
  struct hostent *hp;
  
  server_socket = socket(AF_INET, SOCK_DGRAM, 0);
  if(server_socket < 0){
    perror("opening datagram socket");
    return (-1);
  }

  server.sin_family = AF_INET;
  server.sin_addr.s_addr = INADDR_ANY;
  server.sin_port = htons(5000);
  serv_len = sizeof(server);

  if(bind(server_socket, (struct sockaddr *) &server, serv_len) < 0){
    perror("bind");
    return (-1);    
  }
  
  if(getsockname(server_socket, (struct sockaddr *) &server, &serv_len) < 0){
    perror("getting socket name");
    return (-1);
  }
 
  ext_client.sin_family  = AF_INET;
  ext_client.sin_port = htons(5000);
  cli_len = sizeof(ext_client);  

  if(recvfrom(server_socket, response_buffer, 1024, 0, (struct sockaddr *) &ext_client, &cli_len) > 0) 
    hp = gethostbyaddr((char *) &ext_client.sin_addr, sizeof(ext_client.sin_addr), AF_INET);
  
  printf("The client %s has established a connection with us\n", hp->h_name);
  fflush(stdout);
  
  if(sendto(server_socket, SERVER_RESPONSE, sizeof(SERVER_RESPONSE), 0, (struct sockaddr *) &ext_client, cli_len) < 0) {
    message("Error responding to the client.");
    return (-1);
  }

  message(response_buffer);
  return 0;
}  

int receive_audio()
{
  int buffer_received = 0, data_received = 0, received = 0, nleft = compressed_size;
	
  while(nleft > 0) {
    data_received = recvfrom(server_socket, &receive_buffer[received], nleft, 0, (struct sockaddr *) &ext_client, &cli_len);
    if(data_received < 0){ 
      message("Connection closed by foreign host.");
      return (-1);
    }
    nleft -= data_received;
    received+= data_received;      
  }
  enqueue(&headPtr, &tailPtr);
  
  return 0;
}

int transmit_audio()
{
  int data_transmit = 0, nleft = compressed_size, transmit = 0;
  
  while(nleft > 0) {
    data_transmit = sendto(server_socket, &transmit_buffer[transmit], compressed_size, 0, (struct sockaddr *) &ext_client, serv_len);
    if(data_transmit < 0) {
       message("Connection closed by foreign host.");  
       return (-1);
    }
    nleft-=data_transmit;
    transmit+=data_transmit;
   }

 return 0;
}

main()
{        
  int receive = 0, maxfdp1 = 0, samples_read = 0, x, stop = 0, client_request = 0, data_rceived, request = 0, key_pressed = 0;  
  char answer;
  fd_set readfds;
  struct timeval timeout;   
  struct hostent *gethostbyaddr();
  struct hostent *host;
  int play_size = 0;

  if((Init_Soundcard()) == -1) {
    message("Make sure that the sound card is configured. Refer to Sound-HOWTO for more info.");
    message("Terminating..");
    return (-1);
  }
  
  if((network_connection()) == -1){
    message("Terminating..");
    return (-1);
  }
  maxfdp1 = server_socket + 1;
  timeout.tv_sec = 0;
  timeout.tv_usec = 0;
  FD_ZERO(&readfds);
  FD_SET(server_socket,&readfds);
  
  if((threshold()) < 0) {
    message("Terminating..");
    return(-1);
  }
  message("Please begin the conversation. Click on HANGUP to terminate."); 
  
  for( ; ; ) { 
    receive = 0;
    FD_ZERO(&readfds);
    FD_SET(server_socket,&readfds);
    receive = select(maxfdp1, &readfds, (fd_set *)0, (fd_set *)0, &timeout);
    while(receive) {
      if((receive_audio()) == -1){
	message("Terminating..");
	return (-1);
      }
      receive = 0;
      FD_ZERO(&readfds);
      FD_SET(server_socket,&readfds);
      receive = select(maxfdp1, &readfds, (fd_set *)0, (fd_set *)0, &timeout);
    }
    if(audio_rd == 0) {
      RECEIVING = FALSE;
      TRANSMITTING = TRUE;
      if((open_recording()) == -1) {
	message("Terminating..");
	return (-1);
      }
    }
    samples_read = read(audio_rd, record_buffer, FRAGMENT);
    while((squelch_buffer(FRAGMENT)) == 0) {
      adpcm_coder(record_buffer, compressed_buffer, FRAGMENT, &code);
      memcpy(transmit_buffer, compressed_buffer, compressed_size);
      if((transmit_audio()) == -1) {
	message("Terminating..");
	return (-1);
      }
      samples_read = read(audio_rd, record_buffer, FRAGMENT);
     
    }
    
    FD_ZERO(&readfds);
    FD_SET(0,&readfds);
    key_pressed = select(1, &readfds, (fd_set *)0, (fd_set *)0, &timeout);
    if(key_pressed) {
      close(client_socket);
      close(audio_rd);
      close(audio_wr);
      message("Thank you for using the E-Phone.");
      exit(1);
    }
    
    if(headPtr != NULL) {
      close(audio_rd);
      audio_rd = 0;
      RECEIVING = TRUE;
      TRANSMITTING = FALSE;
      if((open_playback()) == -1) {
	message("Terminating..");
	return(-1);
      }
      while(headPtr != NULL) {
	dequeue(&headPtr, &tailPtr); 
	adpcm_decoder(compressed_buffer,play_buffer, FRAGMENT, &decode);
	write(audio_wr, play_buffer, FRAGMENT);
	receive = 0;
	FD_ZERO(&readfds);
	FD_SET(server_socket,&readfds);
	receive = select(maxfdp1, &readfds, (fd_set *)0, (fd_set *)0, &timeout);
	if(receive) {
	  if((receive_audio()) == -1){
	    message("Terminating..");
	    return (-1);
	  }
	}
      }
      close(audio_wr);
    }
  }
  return 0;
}
          















































































































